@@ -37,6 +37,8 @@
 
 #define N_SCREEN_CONNECTIONS 11
 
+#define NETK_PAGER_GET_PRIVATE(o)    (G_TYPE_INSTANCE_GET_PRIVATE ((o), NETK_TYPE_PAGER, NetkPagerPrivate))
+
 struct _NetkPagerPrivate
 {
     NetkScreen *screen;
@@ -44,6 +46,7 @@ struct _NetkPagerPrivate
     int n_rows;                 /* really columns for vertical orientation */
     NetkPagerDisplayMode display_mode;
     gboolean show_all_workspaces;
+    gboolean workspace_scrolling;
     GtkShadowType shadow_type;
 
     GtkOrientation orientation;
@@ -64,6 +67,8 @@ struct _NetkPagerPrivate
     GdkPixbuf *bg_cache;
 
     int layout_manager_token;
+    int ws_activate;
+    NetkWorkspace *ws_to_activate;
 };
 
 #define POINT_IN_RECT(xcoord, ycoord, rect) \
@@ -110,7 +115,7 @@ static void netk_pager_clear_drag (NetkPager * pager);
 static GdkPixbuf *netk_pager_get_background (NetkPager * pager, int width,
                                              int height);
 
-static gpointer parent_class;
+static GObjectClass *parent_class;
 
 GType
 netk_pager_get_type (void)
@@ -131,6 +136,7 @@ netk_pager_get_type (void)
             sizeof (NetkPager),
             0,                  /* n_preallocs */
             (GInstanceInitFunc) netk_pager_init,
+            NULL		/* value_table */
         };
 
         object_type =
@@ -144,11 +150,12 @@ netk_pager_get_type (void)
 static void
 netk_pager_init (NetkPager * pager)
 {
-    pager->priv = g_new0 (NetkPagerPrivate, 1);
-
+    pager->priv = NETK_PAGER_GET_PRIVATE (pager);
+    
     pager->priv->n_rows = 1;
     pager->priv->display_mode = NETK_PAGER_DISPLAY_CONTENT;
     pager->priv->show_all_workspaces = TRUE;
+    pager->priv->workspace_scrolling = TRUE;
     pager->priv->shadow_type = GTK_SHADOW_NONE;
     pager->priv->orientation = GTK_ORIENTATION_HORIZONTAL;
     pager->priv->workspace_size = 48;
@@ -163,6 +170,8 @@ netk_pager_class_init (NetkPagerClass * klass)
 {
     GObjectClass *object_class = G_OBJECT_CLASS (klass);
     GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
+    
+    g_type_class_add_private (klass, sizeof (NetkPagerPrivate));
 
     parent_class = g_type_class_peek_parent (klass);
 
@@ -195,8 +204,6 @@ netk_pager_finalize (GObject * object)
         pager->priv->bg_cache = NULL;
     }
 
-    g_free (pager->priv);
-
     G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -343,7 +350,7 @@ netk_pager_size_request (GtkWidget * widget, GtkRequisition * requisition)
                 other_dimension_size = MAX (other_dimension_size, w);
             }
 
-            g_object_unref (layout);
+            g_object_unref (G_OBJECT (layout));
 
             other_dimension_size += 2;
         }
@@ -925,7 +932,7 @@ netk_pager_draw_workspace (NetkPager * pager,
                          rect->x + (rect->width - w) / 2,
                          rect->y + (rect->height - h) / 2, layout);
 
-        g_object_unref (layout);
+        g_object_unref (G_OBJECT (layout));
     }
 }
 
@@ -1161,11 +1168,11 @@ netk_pager_button_release (GtkWidget * widget, GdkEventButton * event)
 
             if (space)
             {
-                netk_window_move_to_workspace (pager->priv->drag_window,
-                                               space);
-                if (space ==
-                    netk_screen_get_active_workspace (pager->priv->screen))
+                netk_window_move_to_workspace (pager->priv->drag_window, space);
+                if (space == netk_screen_get_active_workspace (pager->priv->screen))
+                {
                     netk_window_activate (pager->priv->drag_window);
+                }
             }
         }
 
@@ -1179,34 +1186,34 @@ netk_pager_button_release (GtkWidget * widget, GdkEventButton * event)
 
         if (i >= 0 && (space = netk_screen_get_workspace (pager->priv->screen, i)))
         {
-	    int screen_width, screen_height;
-	    
-	    /* Don't switch the desktop if we're already there */		
+            int screen_width, screen_height;
+            
+            /* Don't switch the desktop if we're already there */               
             if (space != netk_screen_get_active_workspace (pager->priv->screen))
                 netk_workspace_activate (space);
+                
+            /* EWMH only lets us move the viewport for the active workspace,
+             * but we just go ahead and hackily assume that the activate
+             * just above takes effect prior to moving the viewport
+             */
 
-                /* EWMH only lets us move the viewport for the active workspace,
-                 * but we just go ahead and hackily assume that the activate
-                 * just above takes effect prior to moving the viewport
-                 */
+            /* Transform the pointer location to viewport origin, assuming
+             * that we want the nearest "regular" viewport containing the
+             * pointer.
+             */                                
+            screen_width  = netk_screen_get_width  (pager->priv->screen);
+            screen_height = netk_screen_get_height (pager->priv->screen);
+            viewport_x = (viewport_x / screen_width)  * screen_width;
+            viewport_y = (viewport_y / screen_height) * screen_height;
 
-            	/* Transform the pointer location to viewport origin, assuming
-	         * that we want the nearest "regular" viewport containing the
-		 * pointer.
-		 */				   
-		screen_width  = netk_screen_get_width  (pager->priv->screen);
-	        screen_height = netk_screen_get_height (pager->priv->screen);
-		viewport_x = (viewport_x / screen_width)  * screen_width;
-		viewport_y = (viewport_y / screen_height) * screen_height;
-				
-		if (netk_workspace_get_viewport_x (space) != viewport_x ||
-	            netk_workspace_get_viewport_y (space) != viewport_y)
-		  netk_screen_move_viewport (pager->priv->screen, viewport_x, viewport_y);
+            if (netk_workspace_get_viewport_x (space) != viewport_x ||
+                netk_workspace_get_viewport_y (space) != viewport_y)
+              netk_screen_move_viewport (pager->priv->screen, viewport_x, viewport_y);
 
-		if ((space == netk_screen_get_active_workspace (pager->priv->screen))
-	             && (pager->priv->drag_window))
-		  netk_window_activate (pager->priv->drag_window);
-	}
+            if ((space == netk_screen_get_active_workspace (pager->priv->screen))
+                 && (pager->priv->drag_window))
+              netk_window_activate (pager->priv->drag_window);
+        }
 
         if (pager->priv->drag_window)
             netk_pager_clear_drag (pager);
@@ -1228,6 +1235,10 @@ netk_pager_scroll_event (GtkWidget * widget, GdkEventScroll * event)
     g_return_val_if_fail (widget != NULL, FALSE);
 
     pager = NETK_PAGER (widget);
+    
+    if (!pager->priv->workspace_scrolling)
+	return FALSE;
+    
     n = netk_screen_get_workspace_count (pager->priv->screen);
     active =
         netk_workspace_get_number (netk_screen_get_active_workspace
@@ -1278,6 +1289,67 @@ netk_pager_focus (GtkWidget * widget, GtkDirectionType direction)
     return GTK_WIDGET_CLASS (parent_class)->focus (widget, direction);
 }
 
+/* dragging something over the pager */
+static gboolean
+netk_pager_drag_timeout (NetkPager *pager)
+{
+    netk_workspace_activate (pager->priv->ws_to_activate);
+    pager->priv->ws_activate = 0;
+    pager->priv->ws_to_activate = NULL;
+    return FALSE;
+}
+
+static void
+netk_pager_drag_leave (GtkWidget * widget, GdkDragContext * context,
+                       guint time, gpointer data)
+{
+    NetkPager *pager = NETK_PAGER (widget);
+
+    if (pager->priv->ws_activate != 0)
+    {
+        g_source_remove (pager->priv->ws_activate);
+        pager->priv->ws_activate = 0;
+        pager->priv->ws_to_activate = NULL;
+    }
+}
+
+static gboolean
+netk_pager_drag_motion (GtkWidget * widget, GdkDragContext * context, gint x,
+                        gint y, guint time, gpointer data)
+{
+    NetkPager *pager = NETK_PAGER (widget);
+    NetkWorkspace *current_ws, *new_ws;
+    int num;
+
+    current_ws = netk_screen_get_active_workspace (pager->priv->screen);
+
+    num = workspace_at_point (pager, x, y, NULL, NULL);
+    
+    if (num != -1) 
+    {
+        new_ws = netk_screen_get_workspace (pager->priv->screen, num);
+        if (new_ws != current_ws && new_ws != pager->priv->ws_to_activate)
+        {
+            if (pager->priv->ws_activate != 0)
+            {
+                g_source_remove (pager->priv->ws_activate);
+            }
+            pager->priv->ws_to_activate = new_ws;
+            pager->priv->ws_activate = 
+                g_timeout_add (1000, (GSourceFunc)netk_pager_drag_timeout, pager);
+        }
+    }
+    else
+    {
+        g_source_remove (pager->priv->ws_activate);
+        pager->priv->ws_activate = 0;
+        pager->priv->ws_to_activate = NULL;
+    }
+    gdk_drag_status (context, 0, time);
+
+    return TRUE;
+}
+
 void
 netk_pager_set_screen (NetkPager * pager, NetkScreen * screen)
 {
@@ -1297,6 +1369,12 @@ netk_pager_new (NetkScreen * screen)
 
     pager = g_object_new (NETK_TYPE_PAGER, NULL);
 
+    gtk_drag_dest_set (GTK_WIDGET (pager), 0, NULL, 0, 0);
+    g_signal_connect (pager, "drag_motion",
+                      G_CALLBACK (netk_pager_drag_motion), NULL);
+    g_signal_connect (pager, "drag_leave",
+                      G_CALLBACK (netk_pager_drag_leave), NULL);
+
     netk_pager_connect_screen (pager, screen);
 
     return GTK_WIDGET (pager);
@@ -1353,7 +1431,7 @@ netk_pager_set_n_rows (NetkPager * pager, int n_rows)
 {
     g_return_if_fail (NETK_IS_PAGER (pager));
 
-    if (pager->priv->n_rows == n_rows)
+    if (pager->priv->n_rows == n_rows || n_rows <= 0)
         return;
 
     pager->priv->n_rows = n_rows;
@@ -1388,6 +1466,18 @@ netk_pager_set_show_all (NetkPager * pager, gboolean show_all_workspaces)
     gtk_widget_queue_resize (GTK_WIDGET (pager));
 }
 
+void
+netk_pager_set_workspace_scrolling (NetkPager * pager, gboolean workspace_scrolling)
+{
+    g_return_if_fail (NETK_IS_PAGER (pager));
+
+    if (pager->priv->workspace_scrolling == workspace_scrolling)
+        return;
+
+    pager->priv->workspace_scrolling = workspace_scrolling;
+    gtk_widget_queue_resize (GTK_WIDGET (pager));
+}
+
 void
 netk_pager_set_shadow_type (NetkPager * pager, GtkShadowType shadow_type)
 {
@@ -1626,8 +1716,7 @@ static void
 netk_pager_connect_window (NetkPager * pager, NetkWindow * window)
 {
     g_signal_connect_object (G_OBJECT (window), "name_changed",
-                             G_CALLBACK (window_name_changed_callback), pager,
-                             0);
+                             G_CALLBACK (window_name_changed_callback), pager, 0);
     g_signal_connect_object (G_OBJECT (window), "state_changed",
                              G_CALLBACK (window_state_changed_callback),
                              pager, 0);
@@ -1635,8 +1724,7 @@ netk_pager_connect_window (NetkPager * pager, NetkWindow * window)
                              G_CALLBACK (window_workspace_changed_callback),
                              pager, 0);
     g_signal_connect_object (G_OBJECT (window), "icon_changed",
-                             G_CALLBACK (window_icon_changed_callback), pager,
-                             0);
+                             G_CALLBACK (window_icon_changed_callback), pager, 0);
     g_signal_connect_object (G_OBJECT (window), "geometry_changed",
                              G_CALLBACK (window_geometry_changed_callback),
                              pager, 0);
