@@ -50,6 +50,8 @@
 #include "netk-xutils.h"
 #include "netk-private.h"
 
+#define NETK_SCREEN_GET_PRIVATE(o)    (G_TYPE_INSTANCE_GET_PRIVATE ((o), NETK_TYPE_SCREEN, NetkScreenPrivate))
+
 static NetkScreen **screens = NULL;
 
 struct _NetkScreenPrivate
@@ -143,7 +145,7 @@ static void emit_background_changed (NetkScreen * screen);
 static void emit_showing_desktop_changed (NetkScreen * screen);
 static void emit_viewports_changed (NetkScreen * screen);
 
-static gpointer parent_class;
+static GObjectClass *parent_class;
 static guint signals[LAST_SIGNAL] = { 0 };
 
 GType
@@ -165,6 +167,7 @@ netk_screen_get_type (void)
             sizeof (NetkScreen),
             0,                  /* n_preallocs */
             (GInstanceInitFunc) netk_screen_init,
+            NULL		/* value_table */
         };
 
         object_type =
@@ -178,14 +181,15 @@ netk_screen_get_type (void)
 static void
 netk_screen_init (NetkScreen * screen)
 {
-    screen->priv = g_new0 (NetkScreenPrivate, 1);
-
+    screen->priv = NETK_SCREEN_GET_PRIVATE (screen);
 }
 
 static void
 netk_screen_class_init (NetkScreenClass * klass)
 {
     GObjectClass *object_class = G_OBJECT_CLASS (klass);
+    
+    g_type_class_add_private (klass, sizeof (NetkScreenPrivate));
 
     p_netk_init ();
 
@@ -315,8 +319,6 @@ netk_screen_finalize (GObject * object)
 
     screens[screen->priv->number] = NULL;
 
-    g_free (screen->priv);
-
     G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -341,7 +343,7 @@ netk_screen_construct (NetkScreen * screen, int number)
     screen->priv->need_update_bg_pixmap = TRUE;
     screen->priv->need_update_showing_desktop = TRUE;
 
-    queue_update (screen);
+    do_update_now (screen);
 }
 
 /**
@@ -1115,7 +1117,7 @@ update_workspace_list (NetkScreen * screen)
     tmp = deleted;
     while (tmp != NULL)
     {
-        g_object_unref (tmp->data);
+        g_object_unref (G_OBJECT (tmp->data));
 
         tmp = tmp->next;
     }
@@ -1557,7 +1559,7 @@ netk_screen_get_height (NetkScreen * screen)
 }
 
 int
-p_netk_screen_get_number (NetkScreen * screen)
+netk_screen_get_number (NetkScreen * screen)
 {
     return screen->priv->number;
 }
