@@ -43,9 +43,11 @@
 #define FALLBACK_NAME "??"
 #define ALL_WORKSPACES (0xFFFFFFFF)
 
+#define NETK_WINDOW_GET_PRIVATE(o)    (G_TYPE_INSTANCE_GET_PRIVATE ((o), NETK_TYPE_WINDOW, NetkWindowPrivate))
+
 static GHashTable *window_hash = NULL;
 
-/* Keep 0-9 in sync with the numbers in the WindowState enum. Yeah I'm
+/* Keep 0-10 in sync with the numbers in the WindowState enum. Yeah I'm
  * a loser.
  */
 #define COMPRESS_STATE(window)                          \
@@ -58,7 +60,8 @@ static GHashTable *window_hash = NULL;
     ((window)->priv->is_sticky           << 6) |        \
     ((window)->priv->is_hidden           << 7) |        \
     ((window)->priv->is_fullscreen       << 8) |        \
-    ((window)->priv->is_urgent           << 9) )
+    ((window)->priv->is_urgent           << 9) |        \
+    ((window)->priv->demands_attention   << 10) )
 
 struct _NetkWindowPrivate
 {
@@ -105,6 +108,7 @@ struct _NetkWindowPrivate
     guint is_hidden:1;
     guint is_fullscreen:1;
     guint is_urgent : 1;
+    guint demands_attention : 1;
 
     /* _NET_WM_STATE_HIDDEN doesn't map directly into an
      * externally-visible state (it determines the WM_STATE
@@ -130,6 +134,9 @@ struct _NetkWindowPrivate
     guint need_update_transient_for:1;
     guint need_update_wmclass : 1;
     guint need_update_wmhints : 1;
+
+    /* if available */
+    char *client_machine;
 };
 
 enum
@@ -171,7 +178,9 @@ static void unqueue_update (NetkWindow * window);
 static void queue_update (NetkWindow * window);
 static void force_update_now (NetkWindow * window);
 
-static gpointer parent_class;
+static NetkWindow *find_last_transient_for (GList * windows, Window xwindow);
+
+static GObjectClass *parent_class;
 static guint signals[LAST_SIGNAL] = { 0 };
 
 GType
@@ -193,6 +202,7 @@ netk_window_get_type (void)
             sizeof (NetkWindow),
             0,                  /* n_preallocs */
             (GInstanceInitFunc) netk_window_init,
+            NULL                /* value_table */
         };
 
         object_type =
@@ -206,10 +216,11 @@ netk_window_get_type (void)
 static void
 netk_window_init (NetkWindow * window)
 {
-    window->priv = g_new0 (NetkWindowPrivate, 1);
-
+    window->priv = NETK_WINDOW_GET_PRIVATE (window);
+    
     window->priv->name = g_strdup (FALLBACK_NAME);;
     window->priv->icon_name = NULL;
+    window->priv->client_machine = NULL;
     window->priv->workspace = ALL_WORKSPACES;
 
     window->priv->icon_cache = p_netk_icon_cache_new ();
@@ -219,6 +230,8 @@ static void
 netk_window_class_init (NetkWindowClass * klass)
 {
     GObjectClass *object_class = G_OBJECT_CLASS (klass);
+    
+    g_type_class_add_private (klass, sizeof (NetkWindowPrivate));
 
     parent_class = g_type_class_peek_parent (klass);
 
@@ -283,6 +296,7 @@ netk_window_finalize (GObject * object)
 
     p_netk_icon_cache_free (window->priv->icon_cache);
 
+    g_free (window->priv->client_machine);
     g_free (window->priv->name);
     g_free (window->priv->icon_name);
     g_free (window->priv->session_id);
@@ -292,8 +306,6 @@ netk_window_finalize (GObject * object)
     if (window->priv->app)
         g_object_unref (G_OBJECT (window->priv->app));
 
-    g_free (window->priv);
-
     G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -363,6 +375,8 @@ p_netk_window_create (Window xwindow, NetkScreen * screen)
     window->priv->session_id = p_netk_get_session_id (window->priv->xwindow);
 
     window->priv->pid = p_netk_get_pid (window->priv->xwindow);
+    
+    window->priv->client_machine = p_netk_get_client_machine (window->priv->xwindow);
 
     p_netk_get_window_geometry (NETK_SCREEN_XSCREEN (window->priv->screen),
                                 xwindow, &window->priv->x, &window->priv->y,
@@ -506,6 +520,20 @@ netk_window_get_session_id_utf8 (NetkWindow * window)
     return window->priv->session_id_utf8;
 }
 
+/**
+ * netk_window_get_pid:
+ * @window: a #NetkWindow
+ *
+ * Gets the client machine FQDN of @window, if available.
+ *
+ * Return value: FQDN of machine of @window, or NULL if none available
+ **/
+const char *
+netk_window_get_client_machine (NetkWindow * window)
+{
+	return window->priv->client_machine;
+}
+
 /**
  * netk_window_get_pid:
  * @window: a #NetkWindow
@@ -637,14 +665,6 @@ netk_window_is_skip_tasklist (NetkWindow * window)
     return window->priv->skip_taskbar;
 }
 
-gboolean
-netk_window_is_fullscreen (NetkWindow * window)
-{
-    g_return_val_if_fail (NETK_IS_WINDOW (window), FALSE);
-
-    return window->priv->is_fullscreen;
-}
-
 void
 netk_window_set_skip_tasklist (NetkWindow * window, gboolean skip)
 {
@@ -654,6 +674,14 @@ netk_window_set_skip_tasklist (NetkWindow * window, gboolean skip)
                          p_netk_atom_get ("_NET_WM_STATE_SKIP_TASKBAR"), 0);
 }
 
+gboolean
+netk_window_is_fullscreen (NetkWindow * window)
+{
+    g_return_val_if_fail (NETK_IS_WINDOW (window), FALSE);
+
+    return window->priv->is_fullscreen;
+}
+
 void
 netk_window_set_fullscreen (NetkWindow * window, gboolean fullscreen)
 {
@@ -664,6 +692,67 @@ netk_window_set_fullscreen (NetkWindow * window, gboolean fullscreen)
                         p_netk_atom_get ("_NET_WM_STATE_FULLSCREEN"), 0);
 }
 
+/**
+ * netk_window_demands_attention:
+ * @window: a #NetkWindow
+ *
+ * If the window demands attention returns %TRUE. This state may change
+ * anytime a state_changed signal gets emitted, as it can depend on flags
+ * such as the demands_attention and is_urgent hints.
+ *
+ * Return value: %TRUE if window demands attention
+ **/
+gboolean
+netk_window_demands_attention (NetkWindow * window)
+{
+    g_return_val_if_fail (NETK_IS_WINDOW (window), FALSE);
+
+    return window->priv->demands_attention || window->priv->is_urgent;
+}
+
+/* Return whether one of the transients of @window demands attention */
+static gboolean
+transient_demands_attention (NetkWindow * window)
+{
+    GList *windows;
+    NetkWindow *transient;
+
+    if (!NETK_IS_WINDOW (window))
+        return FALSE;
+
+    windows = netk_screen_get_windows_stacked (window->priv->screen);
+
+    transient = window;
+    while ((transient =
+            find_last_transient_for (windows, transient->priv->xwindow)))
+    {
+        /* catch transient cycles */
+        if (transient == window)
+            return FALSE;
+
+        if (netk_window_demands_attention (transient))
+            return TRUE;
+    }
+
+    return FALSE;
+}
+
+/**
+ * netk_window_or_transient_demands_attention:
+ * @window: a #NetkWindow
+ *
+ * If the window or one of its transients demands attention returns %TRUE.
+ * This state may change anytime a state_changed signal gets emitted.
+ *
+ * Return value: %TRUE if window or one of its transients demands attention
+ **/
+gboolean
+netk_window_or_transient_demands_attention (NetkWindow * window)
+{
+    return netk_window_demands_attention (window) ||
+        transient_demands_attention (window);
+}
+
 /**
  * netk_window_is_sticky:
  * @window: a #NetkWindow
@@ -1286,8 +1375,10 @@ p_netk_window_set_class_group (NetkWindow     *window,
 
     if (class_group)
         g_object_ref (G_OBJECT (class_group));
+
     if (window->priv->class_group)
         g_object_unref (G_OBJECT (window->priv->class_group));
+
     window->priv->class_group = class_group;
 }
 
@@ -1304,16 +1395,16 @@ p_netk_window_process_property_notify (NetkWindow * window, XEvent * xevent)
         window->priv->need_update_wm_state = TRUE;
         queue_update (window);
     }
-    else if (xevent->xproperty.atom == XA_WM_NAME
-             || xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_NAME")
-             || xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_VISIBLE_NAME"))
+    else if ((xevent->xproperty.atom == XA_WM_NAME) ||
+             (xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_NAME")) ||
+             (xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_VISIBLE_NAME")))
     {
         window->priv->need_update_name = TRUE;
         queue_update (window);
     }
-    else if (xevent->xproperty.atom == XA_WM_ICON_NAME
-             || xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_ICON_NAME")
-             || xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_VISIBLE_ICON_NAME"))
+    else if ((xevent->xproperty.atom == XA_WM_ICON_NAME) ||
+             (xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_ICON_NAME")) ||
+             (xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_VISIBLE_ICON_NAME")))
     {
         window->priv->need_update_icon_name = TRUE;
         queue_update (window);
@@ -1328,8 +1419,7 @@ p_netk_window_process_property_notify (NetkWindow * window, XEvent * xevent)
         window->priv->need_update_workspace = TRUE;
         queue_update (window);
     }
-    else if (xevent->xproperty.atom ==
-             p_netk_atom_get ("_NET_WM_WINDOW_TYPE"))
+    else if (xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_WINDOW_TYPE"))
     {
         window->priv->need_update_wintype = TRUE;
         queue_update (window);
@@ -1345,9 +1435,8 @@ p_netk_window_process_property_notify (NetkWindow * window, XEvent * xevent)
         window->priv->need_update_wmclass = TRUE;
         queue_update (window);
     }
-    else if (xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_ICON")
-             || xevent->xproperty.atom == p_netk_atom_get ("KWM_WIN_ICON")
-             || xevent->xproperty.atom == p_netk_atom_get ("WM_HINTS"))
+    else if ((xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_ICON")) ||
+             (xevent->xproperty.atom == p_netk_atom_get ("KWM_WIN_ICON")))
     {
         p_netk_icon_cache_property_changed (window->priv->icon_cache,
                                             xevent->xproperty.atom);
@@ -1355,6 +1444,8 @@ p_netk_window_process_property_notify (NetkWindow * window, XEvent * xevent)
     }
     else if (xevent->xproperty.atom == p_netk_atom_get ("WM_HINTS"))
     {
+        p_netk_icon_cache_property_changed (window->priv->icon_cache,
+                                            xevent->xproperty.atom);
         window->priv->need_update_wmhints = TRUE;
         queue_update (window);
     }
@@ -1428,6 +1519,7 @@ update_state (NetkWindow * window)
         window->priv->skip_taskbar = FALSE;
         window->priv->skip_pager = FALSE;
         window->priv->net_wm_state_hidden = FALSE;
+        window->priv->demands_attention = FALSE;
 
         atoms = NULL;
         n_atoms = 0;
@@ -1454,6 +1546,8 @@ update_state (NetkWindow * window)
                 window->priv->skip_taskbar = TRUE;
             else if (atoms[i] == p_netk_atom_get ("_NET_WM_STATE_SKIP_PAGER"))
                 window->priv->skip_pager = TRUE;
+            else if (atoms[i] == p_netk_atom_get ("_NET_WM_STATE_DEMANDS_ATTENTION"))
+                window->priv->demands_attention = TRUE;
 
             ++i;
         }
@@ -1788,14 +1882,18 @@ update_wmclass (NetkWindow *window)
 static void
 update_wmhints (NetkWindow *window)
 {
-    gboolean retval;
-    
     if (!window->priv->need_update_wmhints)
       return;
 
     window->priv->need_update_wmhints = FALSE;
-    retval = p_netk_read_hints(window->priv->xwindow);
-    window->priv->is_urgent = (retval ? 1 : 0);
+    if (p_netk_is_urgent (window->priv->xwindow))
+    {
+        window->priv->is_urgent = TRUE;
+    }
+    else
+    {
+        window->priv->is_urgent = FALSE;
+    }
 }
 
 static void
@@ -1857,6 +1955,7 @@ force_update_now (NetkWindow * window)
                                  */
     update_workspace (window);  /* emits signals */
     update_actions (window);
+    
 
     get_icons (window);
 
